const {DocumentClient, UriFactory} = require('documentdb')
const { Observable } = require('rxjs')

const connect = (config) => new Promise((resolve, reject) => {
  const client = new DocumentClient(config.host, {masterKey: config.key})
  const url = UriFactory.createDocumentCollectionUri(
    config.database,
    config.collection
  )
  resolve({
    client,
    url,
    close: () => {}
  })
})

const query = (connection, sql) => Observable.create((observer) => {
  const q = connection.client.queryDocuments(
    connection.url,
    sql,
    {enableCrossPartitionQuery: true}
  )
  var abort = false
  const poll = () =>
    q.nextItem((err, record) => {
      if (err) {

        observer.error(err)
      } else if (record) {
        observer.next(record)
        if (!abort) poll()
      } else {
        observer.complete()
      }
    })
  poll()
  return () => (abort = true)
})

const queryFactory = (config) => (line) => Observable.fromPromise(
  connect(config).then((conn) =>
    query(conn, line)
      .do(null, null, conn.close.bind(conn))
  )
).concatAll()

const createParameter = (name, value) => ({
  name,
  value,
  toString: () => typeof value === 'string' ? `'${value.split("'").join("''")}'`
    : value instanceof Date ? `'${value.toISOString()}'`
    : typeof value === 'object' ? JSON.stringify(value)
    : value === null ? 'NULL'
    : value === undefined ? 'NULL'
    : typeof value !== 'number' ? '[Unknown]'
    : value.toString()
})

const raw = (string) => new ''.constructor(string)

function sql (template) {
  var sqlParts = [template[0]]
  var stringParts = [template[0]]
  var parameters = []
  for (var i = 1; i < arguments.length; i++) {
    var pname = `@P${i}`
    var arg = arguments[i]
    if (typeof arg === 'object' && !(arg instanceof Date)) {
      sqlParts.push(arg, template[i])
      stringParts.push(arg, template[i])
    } else {
      var param = createParameter(pname, arguments[i])
      parameters.push(param)
      sqlParts.push(pname, template[i])
      stringParts.push(param, template[i])
    }
  }
  return {
    query: sqlParts.join(''),
    parameters,
    toString: () => stringParts.join('')
  }
}

const stream = (connection, config) => {
  config = config || {}
  var timestamp = config.timestamp || 0
  var ignoreIds = config.ignoreIds || {}
  var interval = isFinite(config.interval) ? config.interval : 1000
  const startTime = new Date().getTime()
  return query(connection, sql`
    select ${raw(config.limit ? `top ${config.limit}` : '')} *
    from c
    where c._ts >= ${config.timestamp || 0}
    order by c._ts asc
  `)
    .filter(c => !ignoreIds[c._rid])
    .do(c => {
      if (timestamp !== c._ts) {
        timestamp = c._ts
        ignoreIds = {}
        ignoreIds[c._rid] = true
      } else {
        ignoreIds[c._rid] = true
      }
    })
    .concat(Observable.create(observer => {
      const waitFor = Math.max(0, interval - (new Date().getTime() - startTime))
      setTimeout(() => {
        observer.next(stream(connection, Object.assign({}, config, {timestamp, ignoreIds, interval})))
        observer.complete()
      }, waitFor)
    }).mergeAll())
}

const replace = (connection, document) => new Promise((resolve, reject) => {
  connection.client.replaceDocument(
    document._self,
    document,
    err => err ? reject(err) : resolve()
  )
})

const upsert = (connection, document) => new Promise((resolve, reject) => {
  connection.client.upsertDocument(
    connection.url,
    document,
    err => err ? reject(err) : resolve()
  )
})

const replaceStream = (connection, stream) =>
  stream
    .map(document => Observable.create(observer => {
      replace(connection, document)
        .then(
          () => {
            observer.next(document)
            observer.complete()
          },
          err => observer.error(err)
        )
    }))
    .concatAll()

const upsertStream = (connection, stream) =>
  stream
    .map(document => Observable.create(observer => {
      upsert(connection, document)
        .then(
          () => {
            observer.next(document)
            observer.complete()
          },
          err => observer.error(err)
        )
    }))
    .concatAll()

module.exports = {
  connect, query, queryFactory, sql, raw, stream, replace, replaceStream, upsertStream
}
